# -*- coding: utf-8 -*-
#
# This file is part of the imio.scan_helpers distribution (https://github.com/IMIO/imio.scan_helpers).
# Copyright (c) 2023 IMIO
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
from config import COPY_BAT_NAME
from config import DOWNLOAD_DIR
from config import get_bundle_dir
from config import GITHUB_REPO
from config import IS_PROD
from config import MAIN_BACKUP_DIR
from config import MAIN_EXE_NAME
from config import PROFILES_DIRS
from logger import close_logger
from logger import log
from requests.exceptions import ConnectionError
from requests.exceptions import RequestException
from requests.exceptions import Timeout

import json
import os
import re
import requests
import shutil
import subprocess
import sys
import traceback
import zipfile


def copy_sub_files(src_dir, dest_dir, files=()):
    """Copy content from src to dest"""
    for item in os.listdir(src_dir):
        if files and item not in files:
            continue
        s = os.path.join(src_dir, item)
        d = os.path.join(dest_dir, item)
        log.info(f'Copying "{s}" to "{dest_dir}"')
        if os.path.isdir(s):
            if os.path.exists(d):
                shutil.rmtree(d)
            shutil.copytree(s, d)
        else:
            shutil.copy2(s, d)


def copy_release_files_and_restart(src_dir, dest_dir):
    """Will create a bat to copy files after main process has ended and restart the main process without upgrade"""
    exe_path = os.path.join(dest_dir, f"{MAIN_EXE_NAME}.exe")
    script_path = os.path.join(dest_dir, COPY_BAT_NAME)

    with open(script_path, "w") as script:
        script.write("@echo off\n")
        script.write(f'echo Copying "{src_dir}" files to "{dest_dir}""\n')
        script.write("timeout /t 3\n")  # waits for main script to end
        script.write(f'xcopy /s /e /h /r /y /q "{src_dir}\\*" "{dest_dir}"\n')
        script.write(f'start "" "{exe_path}" -nu\n')
        script.write(f'rmdir /s /q "{src_dir}"\n')
        # script.write(f'del "{script_path}"\n')

    if IS_PROD:
        subprocess.Popen(["cmd", "/c", script_path])


def download_update(url, download_path):
    """Download github zip file"""
    response = requests.get(url, stream=True)
    response.raise_for_status()
    with open(download_path, "wb") as file:
        for chunk in response.iter_content(chunk_size=8192):
            file.write(chunk)


def exception_infos(ex):
    """Get exception infos"""
    tb_list = traceback.extract_tb(ex.__traceback__)
    filename, line_number, func_name, text = tb_list[-1]
    return f"in {filename} at line {line_number}: '{ex}'"


def get_dated_backup_dir(backup_dir, day):
    """Get dated backup dir"""
    dated_dir = os.path.join(backup_dir, day)
    if not os.path.exists(dated_dir):
        os.makedirs(dated_dir)
    return dated_dir


def get_download_dir_path():
    """Get download dir path for new version zip files"""
    # maybe use tempdir
    # temp_dir = tempfile.gettempdir()
    return os.path.join(get_bundle_dir(), DOWNLOAD_DIR)


def get_last_dated_backup_dir(backup_dir):
    """Get last dated backup dir"""
    subdirs = read_dir(backup_dir, with_path=False, only_folders=True)
    subdirs = [adir for adir in subdirs if re.match(r"^\d{4}-\d{2}-\d{2}$", adir)]
    if subdirs:
        subdirs.sort(reverse=True)
        return os.path.join(backup_dir, subdirs[0])
    return None


def get_latest_release_version(params, release=None):
    """Get GitHub latest or specified release info"""
    if release:
        url = f"https://api.github.com/repos/{GITHUB_REPO}/releases"
        ret = json_request(url, params)
        if not ret:
            return None, None
        for dic in ret:
            if dic["tag_name"] == release:
                url = f"https://api.github.com/repos/{GITHUB_REPO}/releases/{dic['id']}"
                break
        else:
            stop(f"The release with tag '{release}' cannot be found", params=params)
    else:
        url = f"https://api.github.com/repos/{GITHUB_REPO}/releases/latest"
    latest_release = json_request(url, params)
    if not latest_release:
        return None, None
    return latest_release["tag_name"], latest_release["assets"][0]["browser_download_url"]


def get_main_backup_dir(create=True):
    """Get main backup dir"""
    if not os.path.exists(MAIN_BACKUP_DIR):
        if create:
            os.makedirs(MAIN_BACKUP_DIR)
        else:
            return None
    return MAIN_BACKUP_DIR


def get_parameter(params_file, param=None, default=None):
    """Get a specific parameter or the full dic"""
    dic = {}
    if os.path.exists(params_file):
        with open(params_file) as pf:
            dic = json.load(pf)
    if param:
        return dic.get(param, default)
    else:
        return dic


def get_scan_profiles_dir():
    """Get profile dir to read or write"""
    for prof_dir in PROFILES_DIRS:
        if os.path.exists(prof_dir):
            return prof_dir


def json_request(url, params):
    """Simple json request"""
    proxies = {}
    if params.get("PROXY"):
        for protocol in ["http", "https"]:
            if params.get("PROXY_USER") and params.get("PROXY_PWD"):
                proxies[protocol] = f"{params['PROXY_USER']}:{params['PROXY_PWD']}@{params['PROXY']}"
            else:
                proxies[protocol] = params["PROXY"]
    try:
        response = requests.get(url, proxies=proxies)
        response.raise_for_status()
    except requests.exceptions.ProxyError as err:
        send_log_message(f"Cannot request '{url}' : '{err}' with proxies {proxies}", params)
        return {}
    except Exception as err:
        send_log_message(f"Cannot request '{url}' : '{err}'", params)
        return {}
    return response.json()


def read_dir(dirpath, with_path=False, only_folders=False, only_files=False, to_skip=[]):
    """Read the dir and return files"""
    files = []
    for filename in os.listdir(dirpath):
        if filename in to_skip:
            continue
        if only_folders and not os.path.isdir(os.path.join(dirpath, filename)):
            continue
        if only_files and not os.path.isfile(os.path.join(dirpath, filename)):
            continue
        if with_path:
            files.append(os.path.join(dirpath, filename))
        else:
            files.append(filename)
    return files


def send_log_message(message, params, log_method=log.error, level="ERROR"):
    """Send log message to webservice.

    :param message: The message to send
    :param params: The parameters dic containing webservice info
    :param log_method: The log method to use
    :param level: The log level to include: ERROR or INFO
    :return: True if the message was sent successfully, False otherwise
    """
    data = {
        "client_id": params["CLIENT_ID"],
        "hostname": params["hostname"],
        "message": message,
        "level": level,
        "version": params.get("version"),
    }
    if log_method:
        log_method(message)
    headers = {"Content-Type": "application/json", "Accept": "application/json"}
    try:
        response = requests.post(
            params["SERVER_URL"],
            headers=headers,
            data=json.dumps(data),
            auth=("loguser", params["PLONE_PWD"]),
            timeout=30,  # time in seconds for connexion + read
        )
        response.raise_for_status()  # raises exception when status not 200
        return True
    except Timeout:
        log_method("Timeout while sending log message to %s", params["SERVER_URL"])
        return False
    except ConnectionError:
        log_method("Connection error while sending log message to %s", params["SERVER_URL"])
        return False
    except RequestException as e:
        log_method("Failed to send log message to %s: %s", params["SERVER_URL"], str(e))
        return False


def set_parameter(params_file, key, value):
    """Store client_id in file"""
    if key == "CLIENT_ID" and not re.match(r"^0\d{5}$", value):
        stop(f"Given client_id '{value}' not well formed !")
    dic = get_parameter(params_file)
    dic[key] = value
    with open(params_file, "w") as pf:
        json.dump(dic, pf)
    return dic


def stop(msg="", intup=True, params=None, exit_code=1):
    """Stop the script and log the message.

    :param msg: The message to log
    :param intup: Flag to wait for input before closing
    :param params: The parameters dic containing webservice info. If given the message is sent to the webservice
    :param exit_code: The exit code to return
    """
    if msg:
        log.error(msg)
        if params:
            send_log_message(msg, params, log_method=None)
    if intup:
        input("Press Enter to exit...")
    close_logger()
    sys.exit(exit_code)


def unzip_file(zip_path, extract_to):
    """Unzip downloaded archive and delete it"""
    with zipfile.ZipFile(zip_path, "r") as zip_ref:
        zip_ref.extractall(extract_to)
    os.remove(zip_path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This file is part of the imio.scan_helpers distribution (https://github.com/IMIO/imio.scan_helpers).
# Copyright (c) 2023 IMIO
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
from config import get_bundle_dir
from config import PARAMS_FILE_NAME
from config import PROFILES_DIRS
from datetime import datetime
from logger import close_logger
from logger import log
from utils import copy_sub_files
from utils import exception_infos
from utils import get_dated_backup_dir
from utils import get_main_backup_dir
from utils import get_parameter
from utils import get_scan_profiles_dir
from utils import read_dir
from utils import send_log_message
from utils import stop

import argparse
import os


# Argument parsing
parser = argparse.ArgumentParser()
# parser.add_argument("-v", "--version", action="store_true", dest="version", help="Show version")
# parser.add_argument("-r", "--release", dest="release", help="Get this release")
ns = parser.parse_args()

log.info("Starting backup script")
bundle_dir = get_bundle_dir()
params_file = os.path.join(bundle_dir, PARAMS_FILE_NAME)
parameters = get_parameter(params_file)
try:
    main_prof_dir = get_scan_profiles_dir()
    if not os.path.exists(main_prof_dir):
        stop(f"Profiles dir not found: {PROFILES_DIRS}", params=parameters)
    prof_dirs = read_dir(main_prof_dir, with_path=False, only_folders=True)
    if not prof_dirs:
        stop(f"No profiles found in '{main_prof_dir}'", params=parameters)
    main_backup_dir = get_main_backup_dir()
    day = datetime.now().strftime("%Y-%m-%d")
    dated_backup_dir = get_dated_backup_dir(main_backup_dir, day=day)
    copy_sub_files(main_prof_dir, dated_backup_dir, files=prof_dirs)
    input("Press Enter to exit...")
except Exception as ex:
    send_log_message(f"General error in profiles-backup script, {exception_infos(ex)}", parameters)

log.info("Finished backup script")
close_logger()



# -*- coding: utf-8 -*-
# do not use __init__.py because pyinstaller puts it in _internal/__init__/__init__.pyc
# config.pyc is well directly in _internal/
import os
import platform


BUNDLE_NAME = "imio-scan-helpers"
COPY_BAT_NAME = "copy_release_files_and_restart.bat"
DOWNLOAD_DIR = "_downloads"
EXECUTABLE_NAME = "main.exe"
GITHUB_REPO = "IMIO/imio.scan_helpers"
INTERNAL_DIR = "_internal"
MAIN_BACKUP_DIR = "kofax_backup"
MAIN_EXE_NAME = BUNDLE_NAME
PARAMS_FILE_NAME = "configuration.json"
PROFILES_DIRS = [
    ("ProgramData", "Kofax", "Kofax Express 3.3", "Jobs"),
    ("ProgramData", "Kofax", "Kofax Express 3.2", "Jobs"),
    ("ProgramData", "Kofax", "Kofax Express 3.1", "Jobs"),
]
PROFILES_DIRS = [os.path.join(*tup) for tup in PROFILES_DIRS]
SCRIPT_PROFILES_BACKUP_NAME = "profiles-backup"
SCRIPT_PROFILES_RESTORE_NAME = "profiles-restore"
SERVER_URL = "https://scan-logger.imio-app.be/@message-receiver"


if platform.system() == "Windows" and os.getenv("GITHUB_ACTIONS") is None:
    PROFILES_DIRS = [os.path.join("C:\\", path) for path in PROFILES_DIRS]
    MAIN_BACKUP_DIR = os.path.join("C:\\", "imio", MAIN_BACKUP_DIR)
else:
    PROFILES_DIRS = [os.path.join("test_env", path) for path in PROFILES_DIRS]
    MAIN_BACKUP_DIR = os.path.join("test_env", MAIN_BACKUP_DIR)

BUNDLE_DIR = os.path.dirname(__file__)
IS_PROD = False
if os.path.basename(BUNDLE_DIR) == INTERNAL_DIR:
    BUNDLE_DIR = os.path.dirname(BUNDLE_DIR)
    IS_PROD = True


def get_bundle_dir():
    """Get bundle dir in prod or dev environment"""
    if IS_PROD:
        return BUNDLE_DIR
    else:  # dev mode
        root_dir = os.path.dirname(os.path.dirname(BUNDLE_DIR))
        if os.path.exists(os.path.join(root_dir, "dist", BUNDLE_NAME)):
            return os.path.join(root_dir, "dist", BUNDLE_NAME)
        elif os.path.exists(os.path.join(root_dir, "dist")):
            return os.path.join(root_dir, "dist")
        return root_dir


def get_current_version():
    """Get current version. Here to be used in setup.py without namespace problem."""
    if os.path.exists(os.path.join(BUNDLE_DIR, INTERNAL_DIR, "version.txt")):
        v_path = os.path.join(BUNDLE_DIR, INTERNAL_DIR, "version.txt")
    elif os.path.exists(os.path.join(BUNDLE_DIR, "version.txt")):  # dev mode
        v_path = os.path.join(BUNDLE_DIR, "version.txt")
    else:
        return "0.0.0"
    with open(v_path, "r") as file:
        return file.readline().strip()



# setup.py
from imio.scan_helpers.config import get_current_version
from setuptools import find_packages
from setuptools import setup

import os


setup(
    name="imio.scan_helpers",
    version=get_current_version(),
    description="Various script files to handle local scan tool",
    long_description=open("README.rst").read() + "\n" + open("CHANGES.rst").read(),
    author="Stephan Geulette (IMIO)",
    author_email="support@imio.be",
    project_urls={
        "PyPI": "https://pypi.python.org/pypi/imio.scan_helpers",
        "Source": "https://github.com/IMIO/imio.scan_helpers",
    },
    license="GPL version 3",
    keywords="Scan Windows",
    packages=find_packages(exclude=["ez_setup"]),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Microsoft :: Windows :: Windows 10",
        "Operating System :: Microsoft :: Windows :: Windows 11",
    ],
    python_requires=">=3.12",
    install_requires=[
        "pyinstaller",
        "requests",
    ],
)



from unittest.mock import MagicMock
from unittest.mock import patch
from utils import copy_release_files_and_restart
from utils import copy_sub_files
from utils import download_update
from utils import get_dated_backup_dir
from utils import get_download_dir_path
from utils import get_last_dated_backup_dir
from utils import get_latest_release_version
from utils import get_main_backup_dir
from utils import get_parameter
from utils import get_scan_profiles_dir
from utils import read_dir
from utils import send_log_message
from utils import set_parameter

import os
import shutil
import unittest


def p(path):
    return os.path.join(*(path.split("/")))


class TestUtils(unittest.TestCase):
    def test_copy_sub_files(self):
        dated_backup_dir = get_dated_backup_dir(p("test_env/kofax_backup"), day="2000-01-01")
        copy_sub_files(p("test_env/ProgramData/Kofax/Kofax Express 3.2/Jobs"), dated_backup_dir, files=["IMIO ENTRANT"])
        self.assertTrue(os.path.exists(p("test_env/kofax_backup/2000-01-01/IMIO ENTRANT/file1")))
        self.assertFalse(os.path.exists(p("test_env/kofax_backup/2000-01-01/IMIO SORTANT")))
        shutil.rmtree(p("test_env/kofax_backup/2000-01-01"))

    def test_copy_release_files_and_restart(self):
        copy_release_files_and_restart("anything", "test_env")
        self.assertTrue(os.path.exists(p("test_env/copy_release_files_and_restart.bat")))
        with open(p("test_env/copy_release_files_and_restart.bat"), "r") as file:
            content = file.read()
            self.assertIn('xcopy /s /e /h /r /y /q "anything\\*" "test_env"', content)
        os.remove(p("test_env/copy_release_files_and_restart.bat"))

    def test_download_update(self):
        download_update(
            "https://github.com/IMIO/imio.scan_helpers/raw/main/requirements.txt", p("test_env/download.txt")
        )
        self.assertTrue(os.path.exists(p("test_env/download.txt")))
        with open(p("test_env/download.txt"), "r") as file:
            content = file.read()
            self.assertIn("pip-system-certs\npyinstaller", content)
        os.remove(p("test_env/download.txt"))

    def test_get_dated_backup_dir(self):
        self.assertFalse(os.path.exists(p("test_env/kofax_backup/2000-01-01")))
        result = get_dated_backup_dir(p("test_env/kofax_backup"), day="2000-01-01")
        self.assertEqual(result, p("test_env/kofax_backup/2000-01-01"))
        self.assertTrue(os.path.exists(p("test_env/kofax_backup/2000-01-01")))
        shutil.rmtree(p("test_env/kofax_backup/2000-01-01"))

    def test_get_download_dir_path(self):
        self.assertTrue(get_download_dir_path().endswith(p("helpers/_downloads")))

    def test_get_last_dated_backup_dir(self):
        self.assertIsNone(get_last_dated_backup_dir(p("test_env/kofax_backup")))
        os.makedirs(p("test_env/kofax_backup/2000-01-01"))
        os.makedirs(p("test_env/kofax_backup/2000-01-02"))
        self.assertEqual(get_last_dated_backup_dir(p("test_env/kofax_backup")), p("test_env/kofax_backup/2000-01-02"))
        shutil.rmtree(p("test_env/kofax_backup/2000-01-01"))
        shutil.rmtree(p("test_env/kofax_backup/2000-01-02"))

    def test_get_latest_release_version(self):
        tag_name, download_url = get_latest_release_version({}, release="0.2.5")
        self.assertEqual(tag_name, "0.2.5")
        self.assertTrue(download_url.endswith("imio-scan-helpers-0.2.5.zip"))
        tag_name, download_url = get_latest_release_version({})
        self.assertTrue(tag_name >= "0.3.0")

    @patch("utils.json_request")
    def test_get_latest_release_version_bad_request(self, mock_json_request):
        mock_json_request.return_value = {}
        self.assertTupleEqual((None, None), get_latest_release_version({}, release="0.2.5"))

    def test_get_main_backup_dir(self):
        self.assertEqual(get_main_backup_dir(create=False), p("test_env/kofax_backup"))

    def test_get_set_parameter(self):
        self.assertEqual(get_parameter(p("test_env/configuration.json")), {})
        self.assertIsNone(get_parameter(p("test_env/configuration.json"), "a_key"))
        dic = set_parameter(p("test_env/configuration.json"), "a_key", "a_value")
        self.assertTrue(os.path.exists(p("test_env/configuration.json")))
        self.assertEqual(get_parameter(p("test_env/configuration.json"), "a_key"), "a_value")
        self.assertDictEqual(dic, {"a_key": "a_value"})
        os.remove(p("test_env/configuration.json"))

    def test_get_scan_profiles_dir(self):
        self.assertEqual(get_scan_profiles_dir(), p("test_env/ProgramData/Kofax/Kofax Express 3.2/Jobs"))

    def test_read_dir(self):
        self.assertEqual(
            read_dir(p("test_env/ProgramData/Kofax/Kofax Express 3.2/Jobs")), ["IMIO ENTRANT", "IMIO SORTANT"]
        )
        self.assertEqual(
            read_dir(p("test_env/ProgramData/Kofax/Kofax Express 3.2/Jobs"), with_path=True),
            [
                p("test_env/ProgramData/Kofax/Kofax Express 3.2/Jobs/IMIO ENTRANT"),
                p("test_env/ProgramData/Kofax/Kofax Express 3.2/Jobs/IMIO SORTANT"),
            ],
        )
        self.assertEqual(
            read_dir(p("test_env/ProgramData/Kofax/Kofax Express 3.2/Jobs"), only_folders=True),
            ["IMIO ENTRANT", "IMIO SORTANT"],
        )
        self.assertEqual(read_dir(p("test_env/ProgramData/Kofax/Kofax Express 3.2/Jobs"), only_files=True), [])
        self.assertEqual(
            read_dir(p("test_env/ProgramData/Kofax/Kofax Express 3.2/Jobs"), to_skip=["IMIO ENTRANT"]), ["IMIO SORTANT"]
        )

    @patch("requests.post")
    def test_send_log_message(self, mock_post):
        params = {
            "CLIENT_ID": "010001",
            "hostname": "pc1",
            "PLONE_PWD": "password",
            "SERVER_URL": "http://example.com",
        }
        logs = []
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_post.return_value = mock_response
        send_log_message("My message", params, log_method=logs.append, level="ERROR")
        self.assertListEqual(logs, ["My message"])


if __name__ == "__main__":
    unittest.main()
